/*

This program demonstrates how to use Virtual Audio Cable control API.

Can be built using MS VC++ 2005 (compiler version 14.00) and higher.

Don't forget to replace GUIDs in params.h.

To build the executable, use the following command line:

cl vacapi.cpp

You can use DDK build utility or Visual Studio project instead.

*/





#include "prefix.h"



__pragma (warning (disable: 4514 4710))

#if _MSC_VER >= 1900

  __pragma (warning (disable: 5045))

#endif



  __pragma (warning (push))
__pragma (warning (disable: 4201 4619 4668 4820))


#include <stdlib.h>
#include <stdio.h>

#include <windows.h>

__pragma (warning (disable: 4201))

#include <setupapi.h>
#include <winioctl.h>

__pragma (warning (pop))



#include "params.h"



__pragma (warning (push))
__pragma (warning (disable: 4619 4464 4820))

#include "..\vacapi.h"

__pragma (warning (pop))



using namespace VirtualAudioCable;

__pragma (warning (disable: 4482 4711))





/*
######################################################################

Compiler and linker directives
=================================

######################################################################
*/





#pragma comment (lib, "user32.lib")
#pragma comment (lib, "setupapi.lib")





/*
######################################################################

Types/classes
================

######################################################################
*/





typedef UINT LocSize;           // Type of local object size





class Device {

  HDEVINFO DL;
  SP_DEVINFO_DATA DI;

  bool ChangeProperty (DWORD State, DWORD Scope);

  public:

  Device ();
  ~Device ();
    
  void DeleteList ();

  bool Find (PWSTR ForIfPath);
  bool Restart ();

};




class ControlInterface {

  Device Dev;
  HANDLE DriverHandle;

  bool SetNumCables (UINT const N);

  public:

  ControlInterface ();
  ~ControlInterface ();

  bool Open ();
  void Close ();

  bool IoCtl (
    DWORD const CtlCode,
    DriverReq & DR,
    DWORD const ReqSize,
    PVOID const Data2,
    DWORD const Data2Size,
    DWORD * RetSize = nullptr
  ) const;

  bool GetDriverInfo (DriverInfo & DI) const;

  bool SetNumCablesAndRestart (UINT const NumCables);

  bool SetConnSourceLinesMask (UINT const Cable, UINT const Mask, UINT const Flags = 0) const;
  bool SetConnDestLinesMask (UINT const Cable, UINT const Mask) const;

};





/*
######################################################################

Constants
============

######################################################################
*/





struct __declspec (uuid (ProductGuidStr)) ProductGuid;





/*
######################################################################

Variables
============

######################################################################
*/





#if _Debug

  UINT DebugPrintLevel = 9; // Debug importance level

#endif // _Debug





/*
######################################################################

Helper functions
===================

######################################################################
*/





void _cdecl Info (PCTSTR const Format, ...) {

  va_list Args;

  va_start (Args, Format);
  vfwprintf (stdout, Format, Args);
  fputws (L"\n", stdout);
  va_end (Args);
  
}





void _cdecl Error (PCTSTR const Format, ...) {

  fputws (L"***** ", stderr);

  va_list Args;

  va_start (Args, Format);
  vfwprintf (stderr, Format, Args);
  va_end (Args);
  
  fputws (L"\n", stderr);

}





#if _Debug





/*
Debug print with arglist
======================================================================
*/

void Debug (UINT const Level, PCWSTR const Fmt, va_list const List) {

  if (Level >= DebugPrintLevel) {

    if (Level >= 9) {

      fwprintf (stdout, L"***** ");

    }

    vfwprintf (stdout, Fmt, List);

    fputws (L"\n", stdout);

  }

}





#endif // _Debug





/*
Debug print with variable arguments
======================================================================
*/

void _cdecl Debug (UINT const Level, PCWSTR const Fmt, ...) {

#if _Debug

  va_list List;

  if (Level >= DebugPrintLevel) {

    va_start (List, Fmt);
    Debug (Level, Fmt, List);
    va_end (List);

  }

#else // _Debug

  Unreferenced (Level && Fmt);

#endif // _Debug

}





/*
Translate SetupAPI error code from ERROR_XXX to SPAPI_E_XXX
======================================================================

Allows to properly obtain an error description using FormatMessage

*/

void TranslateSetupApiError () {

  DWORD const OldCode = GetLastError ();

  DWORD const Mask = APPLICATION_ERROR_MASK | ERROR_SEVERITY_ERROR;

  if ((OldCode & Mask) == Mask) {

    DWORD const NewCode = DWORD (HRESULT_FROM_SETUPAPI (OldCode));

    if (NewCode != OldCode) {

      Debug (7, L"SetupAPI error code changed from %X to %X", OldCode, NewCode);

      SetLastError (NewCode);

    }

  }

}





/*
Obtain a textual error description
======================================================================
*/

static PCWSTR ErrorText (DWORD const Code) {

  static wchar_t Text [256];

  DWORD const Err = GetLastError ();

  wsprintf (
    Text,
    L"%u/%u - ",
    HRESULT_FACILITY (Code), HRESULT_CODE (Code)
  );

  LocSize const Offset = static_cast <LocSize> (wcslen (Text));

  wchar_t * cp = Text + Offset;
  LocSize const Avail = NumArrayElems (Text) - Offset - 1;

  FormatMessage (
    FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK,
    nullptr,
    Code,
    MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
    cp,
    UINT (Avail),
    nullptr 
  );

  cp += wcslen (cp) - 1;

  if (*cp == ' ') {

    *cp = '\0';

  }

  SetLastError (Err);

  return Text;

}





/*
Obtain a textual description of a last Win32 error
======================================================================
*/

static inline PCWSTR LastErrorText () {

  return ErrorText (GetLastError ());

}





/*
Convert a GUID to a string
======================================================================

GUID string size is 36.

*/

PTSTR GuidText (PTSTR const Buffer, LocSize const MaxLen, GUID const & Guid) {

  static wchar_t InternalBuffer [40];

  PTSTR const ActualBuffer = (Buffer != nullptr)? Buffer : InternalBuffer;
  LocSize const ActualMaxLen = (Buffer != nullptr)? MaxLen : NumArrayElems (InternalBuffer);

  wchar_t Str [40];

  if (&Guid != nullptr) {

    wsprintf (
      Str,
      L"%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
      Guid.Data1,
      Guid.Data2, Guid.Data3,
      Guid.Data4 [0], Guid.Data4 [1],
      Guid.Data4 [2], Guid.Data4 [3], Guid.Data4 [4],
      Guid.Data4 [5], Guid.Data4 [6], Guid.Data4 [7]
    );

  } else {

    wcscpy (Str, L"GUID_NULL");

  }

  LocSize const Len = static_cast <LocSize> (wcslen (Str));
  LocSize const ActualLen = min (Len, ActualMaxLen - 1);

  wcsncpy (ActualBuffer, Str, ActualLen);

  ActualBuffer [ActualLen] = L'\0';

  return ActualBuffer;

}





/*
######################################################################

Device class implementation
======================================================================

######################################################################
*/





/*
Constructor
======================================================================
*/

Device::Device () {

  DL = 0;

}





/*
Destructor
======================================================================
*/

Device::~Device () {

  DeleteList ();

}





/*
Delete device list
======================================================================
*/

void Device::DeleteList () {

  if (DL) {

    Debug (6, L"Deleting device list");

    SetupDiDestroyDeviceInfoList (DL);
    DL = 0;

  }

}





/*
Find the driver
======================================================================
*/

bool Device::Find (PWSTR const ForIfPath) {

  bool Success = false;

  Assert (DL == 0);

  do {

    Debug (
      7,
      L"Searching for VAC driver in class/category %s",
      GuidText (nullptr, 0, __uuidof (ProductGuid))
    );

    // Get a list of devices bound to a given interface

    DL = SetupDiGetClassDevs (
      const_cast <GUID *> (&__uuidof (ProductGuid)),
      0,
      0,
      DIGCF_DEVICEINTERFACE | DIGCF_PRESENT
    );

    if (DL == INVALID_HANDLE_VALUE) {

      TranslateSetupApiError ();

      Error (L"Cannot get class device list (%s)", LastErrorText ());

      DL = 0;

      break;

    }

    SP_DEVICE_INTERFACE_DATA IfData;

    RtlZeroMemory (&IfData, sizeof (IfData));

    IfData.cbSize = sizeof (IfData);

    BOOL Res;

    // Get first interface data

    Res = SetupDiEnumDeviceInterfaces (
      DL, nullptr,
      const_cast <GUID *> (&__uuidof (ProductGuid)), 0, &IfData
    );

    if (!Res) {

      TranslateSetupApiError ();

      if (GetLastError () == ERROR_NO_MORE_ITEMS) {

        Error (L"Driver not found");

      } else {

        Error (L"Cannot get interface data (%s)", LastErrorText ());

      }

      break;

    }

    struct IfDetail : public SP_DEVICE_INTERFACE_DETAIL_DATA {

      BYTE VarPart [MAX_PATH];
    
    } IfDetData;

    RtlZeroMemory (&IfDetData, sizeof (IfDetData));

    IfDetData.cbSize = sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);

    RtlZeroMemory (&DI, sizeof (DI));

    DI.cbSize = sizeof (DI);

    // Get interface details

    Res = SetupDiGetDeviceInterfaceDetail (
      DL,
      &IfData,
      &IfDetData,
      sizeof (IfDetData),
      0,
      &DI
    );

    if (!Res) {

      TranslateSetupApiError ();

      Error (L"Cannot get interface details (%s)", LastErrorText ());

      break;

    }

    Debug (6, L"Driver interface: %s", IfDetData.DevicePath);

    if (ForIfPath) {

      wcscpy (ForIfPath, IfDetData.DevicePath);

    }

    Success = true;

  } OnlyOnce;

  if (!Success) {

    DeleteList ();

  }

  return Success;

}





/*
Change device property
======================================================================
*/

bool Device::ChangeProperty (DWORD const State, DWORD const Scope) {

  SP_PROPCHANGE_PARAMS PCP;

  PCP.ClassInstallHeader.cbSize = sizeof (PCP.ClassInstallHeader);
  PCP.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
  PCP.StateChange = State;
  PCP.Scope = Scope;
  PCP.HwProfile = 0;

  bool Success = false;

  do {

    Debug (6, L"Setting property change parameters");

    if (
      !SetupDiSetClassInstallParams (
        DL,
        &DI,
        &PCP.ClassInstallHeader,
        sizeof (PCP)
      )
    ) {

      TranslateSetupApiError ();

      Error (
        L"Cannot set propchange params: State=%u, Scope=%u (%s)",
        State, Scope, LastErrorText ()
      );

      break;

    }

    Debug (6, L"Issuing property change request");

    if (!SetupDiCallClassInstaller (DIF_PROPERTYCHANGE, DL, &DI)) {

      TranslateSetupApiError ();

      Error (L"Cannot change device property (%s)", LastErrorText ());

      break;

    }

    Success = true;

  } OnlyOnce;

  return Success;

}





/*
Restart device (reload driver)
======================================================================
*/

bool Device::Restart () {

  // Administrator rights are required to restart the driver. To change
  // parameters without driver restart, no special rights are requred.

  Debug (6, L"Restarting driver");

  bool Success = false;

  do {
    
    if (!ChangeProperty (DICS_PROPCHANGE, DICS_FLAG_GLOBAL)) {

      Error (L"Cannot issue property change command (%s)", LastErrorText ());

      break;

    }

    SP_DEVINSTALL_PARAMS DIP;

    DIP.cbSize = sizeof (DIP);

    if (!SetupDiGetDeviceInstallParams (DL, &DI, &DIP)) {

      Error (L"Cannot get device status (%s)", LastErrorText ());

      break;

    }

    if (DIP.Flags & DI_NEEDRESTART) {

      Debug (7, L"Restart is required");

    } else if (DIP.Flags & DI_NEEDREBOOT) {

      Debug (7, L"Reboot is required");

    } else {

      Debug (7, L"Driver restarted");

    }

    Success = true;

  } OnlyOnce;

  return Success;

}





/*
######################################################################

ControlInterface class implementation
======================================================================

######################################################################
*/





/*
Constructor
==============
*/

ControlInterface::ControlInterface () {

  DriverHandle = 0;

}





/*
Destructor
======================================================================
*/

ControlInterface::~ControlInterface () {

  Close ();

}





/*
Open driver's control interface
======================================================================
*/

bool ControlInterface::Open () {

  Assert (DriverHandle == 0);

  bool Success = false;

  do {

    wchar_t ControlInterfacePath [MAX_PATH];

    if (!Dev.Find (ControlInterfacePath)) {

      break;

    }

    // Append the control interface GUID

    wcscat (ControlInterfacePath, L"\\");
    wcscat (ControlInterfacePath, _UT (ApiInterfaceGuidStr));

    Debug (6, L"Opening %s", ControlInterfacePath);

    Assert (DriverHandle == 0);

    DriverHandle = CreateFile (
      ControlInterfacePath, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0
    );

    if (DriverHandle == INVALID_HANDLE_VALUE) {

      DriverHandle = 0;

      Error (L"Cannot open driver (%s)", LastErrorText ());

      break;

    }

    Debug (7, L"Control interface opened");

    Success = true;

  } OnlyOnce;

  return Success;

}





/*
Close driver interface
======================================================================
*/

void ControlInterface::Close () {

  if (DriverHandle) {

    Debug (6, L"Closing driver interface");

    CloseHandle (DriverHandle);
    DriverHandle = 0;

  }

}





/*
Send an IOCTL request to the driver
======================================================================
*/

bool ControlInterface::IoCtl (
  DWORD const CtlCode,
  DriverReq & DR,
  DWORD const ReqSize,
  PVOID const Data2,
  DWORD const Data2Size,
  DWORD * RetSize
) const {

  DWORD LocRetSize;

  bool Success = false;

  do {

    if (!RetSize) {

      RetSize = &LocRetSize;

    }

    Debug (5, L"Issuing IOCTL with CC=%X", CtlCode);

    Assert (DriverHandle != 0);

    DR.ApiVer = ApiVersion;

    BOOL const Res = DeviceIoControl (
      DriverHandle,
      CtlCode,
      const_cast <DriverReq *> (&DR), ReqSize,
      Data2, Data2Size,
      RetSize,
      nullptr
    );


    Debug (4, L"IoCtl returned %u, LastErr=%u", Res, GetLastError ());

    if (Res) {
    
      Success = true;

    } else {

      Debug (9, L"IoCtl failed (%s)", LastErrorText ());

      Success = false;

    }

  } OnlyOnce;

  return Success;

}





/*
Request driver information
======================================================================
*/

bool ControlInterface::GetDriverInfo (DriverInfo & DI) const {

  Debug (5, L"GetDriverInfo");

  DriverInfoReq R;

  bool const Res = IoCtl (IoCtl_Local_GetDriverInfo, R, sizeof (R), &DI, sizeof (DI));

  if (!Res) {

    Error (L"Cannot get driver info (%s)", LastErrorText ());

  }

  return Res;

}





/*
Send number of cables change request
======================================================================
*/

bool ControlInterface::SetNumCables (UINT const N) {

  Debug (7, L"SetNumCables (%u)", N);

  DriverSetNumCablesReq R;

  R.NumCables = N;

  bool const Res = IoCtl (IoCtl_Local_SetNumCables, R, sizeof (R), nullptr, 0);

  if (!Res) {

    Error (L"Cannot set number of cables to %u (%s)", N, LastErrorText ());

  }

  return Res;

}





/*
Set number of cables
======================================================================
*/

bool ControlInterface::SetNumCablesAndRestart (UINT const NumCables) {

  bool Success = false;

  do {

    if (!SetNumCables (NumCables)) {

      break;

    }

    DriverInfo DI;

    if (!GetDriverInfo (DI)) {

      break;

    }

    if (TestBits (DI.Flags, DriverFlags::RestartRequired)) {

      Close ();

      Dev.Restart ();

      Dev.DeleteList ();

      if (!Open ()) {

        break;

      }

    }

    Success = true;

  } OnlyOnce;

  return Success;

}





/*
Set connected source lines mask
======================================================================
*/

bool ControlInterface::SetConnSourceLinesMask (UINT const Cable, UINT const Mask, UINT const Flags) const {

  Debug (7, L"SetConnSourceLinesMask (%#x)", Mask);

  CableSetConnectedSourceLinesMaskExReq R;

  R.CableNum = Cable;
  R.ConnectedSourceLinesMask = Mask;
  R.Flags = Flags;

  bool const Res = IoCtl (
    IoCtl_Local_SetConnectedSourceLinesMask,
    R, (Flags != 0)? sizeof (R) : sizeof (CableSetConnectedSourceLinesMaskReq),
    nullptr, 0
  );

  if (!Res) {

    Error (
      L"Cable %u: cannot set connected source lines mask to %#x (%s)",
      Cable + 1, Mask, LastErrorText ()
    );

  }

  return Res;

}





/*
Set connected destination lines mask
======================================================================
*/

bool ControlInterface::SetConnDestLinesMask (UINT const Cable, UINT const Mask) const {

  Debug (7, L"SetConnDestLinesMask (%#x)", Mask);

  CableSetConnectedDestLinesMaskReq R;

  R.CableNum = Cable;
  R.ConnectedDestLinesMask = Mask;

  bool const Res = IoCtl (IoCtl_Local_SetConnectedDestLinesMask, R, sizeof (R), nullptr, 0);

  if (!Res) {

    Error (
      L"Cable %u: cannot set connected destination lines mask to %#x (%s)",
      Cable + 1, Mask, LastErrorText ()
    );

  }

  return Res;

}





/*
######################################################################

Global functions
======================================================================

######################################################################
*/





/*
Command: get driver version 
======================================================================
*/

UINT Command_Version (UINT const NumArgs, PCTSTR const * const Args) {

  UINT Res = 999;

  ControlInterface API;

  do {

    if (NumArgs != 0) {

      Error (L"Extra parameter in command line - \"%s\"", Args [0]);

      break;

    }

    if (!API.Open ()) {

      break;

    }

    DriverInfo DI;

    if (!API.GetDriverInfo (DI)) {

      break;

    }

    Info (
      L"Driver version: %u.%u.%u.%u, %s, %u cables",
      DI.VerMajor, DI.VerMinor, DI.VerRev, DI.VerBuild,
      TestBits (DI.Flags, DriverFlags::Trial)? L"trial" : L"full",
      DI.NumCables
    );

    Res = 0;

  } OnlyOnce;

  return Res;

}





/*
Command: set number of cables
======================================================================
*/

UINT Command_SetNumberOfCables (UINT const NumArgs, PCTSTR const * const Args) {

  UINT Res = 999;

  ControlInterface API;

  do {

    if (NumArgs > 1) {

      Error (L"Extra parameter in command line - \"%s\"", Args [1]);

      break;

    } else if (NumArgs < 1) {

      Error (L"Number of cables not specified");

      break;

    }

    PTSTR EndPtr;

    UINT const Num = wcstoul (Args [0], &EndPtr, 10);

    Assert (EndPtr != nullptr);

    if (*EndPtr != L'\0' || Num == 0 || Num > 256) {

      Error (L"Invalid number of cables");

      break;

    }

    if (!API.Open ()) {

      break;

    }

    DriverInfo DI;

    if (!API.GetDriverInfo (DI)) {

      break;

    }

    if (Num == DI.NumCables) {

      Info (L"Number of cables already set to %u", Num);

      break;

    }

#ifndef _M_X64

    BOOL Wow64;

    if (!IsWow64Process (GetCurrentProcess (), &Wow64)) {

      Error (L"Cannot query WOW64 status (%s)", LastErrorText ());

      break;

    }

    if (Wow64) {

      Error (L"To set number of cables in 64-bit system, use 64-bit executable");

      break;

    }

#endif

    if (!API.SetNumCablesAndRestart (Num)) {

      break;

    }

    if (!API.GetDriverInfo (DI)) {

      break;

    }

    Info (L"New number of cables: %u", DI.NumCables);

    Res = 0;

  } OnlyOnce;

  return Res;

}





/*
Command: set mask of connected source lines
======================================================================
*/

UINT Command_SetConnSourceLinesMask (UINT const NumArgs, PCTSTR const * const Args) {

  UINT Res = 999;

  ControlInterface API;

  do {

    if (NumArgs > 3) {

      Error (L"Extra parameter in command line - \"%s\"", Args [2]);

      break;

    } else if (NumArgs < 1) {

      Error (L"Cable number not specified");

      break;

    } else if (NumArgs < 2) {

      Error (L"Connected source lines mask not specified");

      break;

    }

    PTSTR EndPtr;

    UINT const Cable = wcstoul (Args [0], &EndPtr, 10);

    Assert (EndPtr != nullptr);

    if (*EndPtr != L'\0' || Cable >= 256) {

      Error (L"Invalid cable number %u", Cable);

      break;

    }

    UINT const Mask = wcstoul (Args [1], &EndPtr, 10);

    Assert (EndPtr != nullptr);

    if (*EndPtr != L'\0' || TestBits (Mask, ~((1u << SourceLine::Types::Total) - 1))) {

      Error (L"Invalid connected source lines mask %#x", Mask);

      break;

    }

    UINT Flags = 0;

    if (NumArgs == 3) {

      Flags = wcstoul (Args [2], &EndPtr, 10);

      Assert (EndPtr != nullptr);

      if (*EndPtr != L'\0' || TestBits (Flags, ~((1u << SourceLine::Types::Total) - 1))) {

        Error (L"Invalid connected source lines flags %#x", Flags);

        break;

      }

    }

    if (!API.Open ()) {

      break;

    }

    if (!API.SetConnSourceLinesMask (Cable, Mask, Flags)) {

      break;

    }

    Info (L"Cable %u: connected source lines mask set to %#x", Cable, Mask);

    Res = 0;

  } OnlyOnce;

  return Res;

}





/*
Command: set mask of connected destination lines
======================================================================
*/

UINT Command_SetConnDestLinesMask (UINT const NumArgs, PCTSTR const * const Args) {

  UINT Res = 999;

  ControlInterface API;

  do {

    if (NumArgs > 2) {

      Error (L"Extra parameter in command line - \"%s\"", Args [2]);

      break;

    } else if (NumArgs < 1) {

      Error (L"Cable number not specified");

      break;

    } else if (NumArgs < 2) {

      Error (L"Connected destination lines mask not specified");

      break;

    }

    PTSTR EndPtr;

    UINT const Cable = wcstoul (Args [0], &EndPtr, 10);

    Assert (EndPtr != nullptr);

    if (*EndPtr != L'\0' || Cable >= 256) {

      Error (L"Invalid cable number %u", Cable);

      break;

    }

    UINT const Mask = wcstoul (Args [1], &EndPtr, 10);

    Assert (EndPtr != nullptr);

    if (*EndPtr != L'\0' || Mask > 1) {

      Error (L"Invalid connected destination lines mask %#x", Mask);

      break;

    }

    if (!API.Open ()) {

      break;

    }

    if (!API.SetConnDestLinesMask (Cable, Mask)) {

      break;

    }

    Info (L"Cable %u: connected destination lines mask set to %#x", Cable, Mask);

    Res = 0;

  } OnlyOnce;

  return Res;

}





/*
Main function
================
*/

int _cdecl wmain (int const ArgC, PCTSTR const ArgV []) {

  UINT Res = 99;

  do {

    if (ArgC < 2) {

      Info (
        L"Virtual Audio Cable 4 API usage demonstration\n"
        L"\n"
        L"Command line: [<options>] <command> [<parameters>]\n"
        L"\n"
        L"Options:\n"
#if _Debug
        L"\n"
        L"-v - verbose.\n"
#endif _Debug
        L"\n"
        L"Commands:\n"
        L"\n"
        L"version - show driver version.\n"
        L"setnumcables <n> - set number of cables.\n"
        L"setconnsrclines <cable> <mask> [<flags>] - set connected destination lines mask.\n"
        L"setconndstlines <cable> <mask> - set connected destination lines mask.\n"
      );

      Res = 1;

      break;

    }



    UINT NumArgs = static_cast <UINT> (ArgC - 1);
    PCTSTR const * Args = ArgV + 1;



#if _Debug

    if (wcscmp (*Args, L"-v") == 0) {

      --NumArgs;
      ++Args;

      DebugPrintLevel = 5;

    }

#endif _Debug



    if (NumArgs == 0) {

      Error (L"Command not specified");

      Res = 2;
      break;

    }



    PCTSTR const Command = *Args++;

    --NumArgs;

    if (wcscmp (Command, L"version") == 0) {

      Res = Command_Version (NumArgs, Args);

    } else if (wcscmp (Command, L"setnumcables") == 0) {

      Res = Command_SetNumberOfCables (NumArgs, Args);

    } else if (wcscmp (Command, L"setconnsrclines") == 0) {

      Res = Command_SetConnSourceLinesMask (NumArgs, Args);

    } else if (wcscmp (Command, L"setconndstlines") == 0) {

      Res = Command_SetConnDestLinesMask (NumArgs, Args);

    } else {

      Error (L"Invalid command \"%s\"", Command);

      Res = 2;
      break;

    }



  } OnlyOnce;

  return static_cast <int> (Res);

}
