#include "common.h"





bool GetTokenInfo (HANDLE const Token, TOKEN_INFORMATION_CLASS const Type, PVOID & Data) {

  Assert (Token != 0);

  bool Success = false;

  {

    DWORD RetSize;

    bool const Res = !!GetTokenInformation (Token, Type, nullptr, 0, &RetSize);

    if (Res || GetLastError() != ERROR_INSUFFICIENT_BUFFER) {

      Error (L"Cannot get size of token info %u (%s)", Type, LastErrorText ());

      goto Exit;

    }

    //Debug (8, L"%u bytes needed for token information", RetSize);

    Assert (RetSize != 0);

    PVOID const xData = new BYTE [RetSize];

    Assert (xData != nullptr);

    if (!GetTokenInformation (Token, Type, xData, RetSize, &RetSize)) {

      Error (L"Cannot get token info %u (%s)", Type, LastErrorText ());

      goto Exit;

    }

    Data = xData;

    Success = true;

  }

Exit:

  return Success;

}





bool GetTokenInfos () {

  HANDLE Token = 0;
  bool Success = false;

  {

    if (!OpenProcessToken (GetCurrentProcess(), TOKEN_READ, &Token)) {

      Error (L"Cannot open process token for read (%s)", LastErrorText ());

      goto Exit;

    }

    Assert (UserToken == nullptr);

    if (!GetTokenInfo (Token, TokenUser, *reinterpret_cast <PVOID *> (&UserToken))) {

      goto Exit;

    }

    Assert (OwnerToken == nullptr);

    if (!GetTokenInfo (Token, TokenOwner, *reinterpret_cast <PVOID *> (&OwnerToken))) {

      goto Exit;

    }

    Success = true;

  }

Exit:

  if (Token != 0) {

    Verify (CloseHandle (Token));
    Token = 0;

  }

  return Success;

}




HRESULT AdjustPrivilege (PCTSTR const PrivilegeName, bool const Enable) {

  HRESULT FinalRes;
  HANDLE Token = 0;

  {

    HRESULT HR;
    BOOL Res;

    Res = OpenProcessToken (GetCurrentProcess (), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &Token);

    if (!Res) {

      FinalRes = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot open process token (%s)", ErrorText (FinalRes));

      Token = 0;
      goto Exit;
  
    } 

    TOKEN_PRIVILEGES Privs;

    Res = LookupPrivilegeValue (nullptr, PrivilegeName, &Privs.Privileges [0].Luid);

    if (!Res) {

      HR = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot lookup value of %s (%s)", PrivilegeName, ErrorText (HR));

      FinalRes = HR;
      goto Exit;
  
    } 

    Privs.PrivilegeCount = 1;
    Privs.Privileges [0].Attributes = static_cast <DWORD> (Enable? SE_PRIVILEGE_ENABLED : 0);

    Res = AdjustTokenPrivileges (Token, false, &Privs, 0, nullptr, nullptr);

    if (!Res) {

      HR = HRESULT_FROM_WIN32 (GetLastError ());

      Trace (9, L"Cannot %sable privilege %s (%s)", Enable? L"en" : L"dis", PrivilegeName, ErrorText (HR));

      FinalRes = HR;
      goto Exit;

    }

    Trace (8, L"Privilege %s %sabled", PrivilegeName, Enable? L"en" : L"dis");

    FinalRes = S_OK;

  }

Exit:

  if (Token != 0) {

    Verify (CloseHandle (Token));
    Token = 0;

  }

  return FinalRes;

}





HRESULT TakeOwnership (SE_OBJECT_TYPE const ObjectType, PCTSTR const ObjectPath) {

  HRESULT FinalRes;

  {

    Trace (7, L"Taking ownership of \"%s\"", ObjectPath);

    HRESULT HR;
    DWORD Res;

    HR = AdjustPrivilege (SE_TAKE_OWNERSHIP_NAME, true);

    if (FAILED (HR)) {

      FinalRes = HR;
      goto Exit; 

    }

    Assert (OwnerToken != nullptr);

    Res = SetNamedSecurityInfo (
      const_cast <PTSTR> (ObjectPath), ObjectType,
      OWNER_SECURITY_INFORMATION, OwnerToken->Owner,
      nullptr, nullptr, nullptr
    );

    if (Res != NO_ERROR) {

      Trace (9, L"Cannot change owner of \"%s\" (%s)", ObjectPath, ErrorText (Res));

      FinalRes = HRESULT_FROM_WIN32 (Res);
      goto Exit;

    }
        
    Trace (8, L"Ownership of \"%s\" is taken", ObjectPath);

    HR = AdjustPrivilege (SE_TAKE_OWNERSHIP_NAME, false);

    if (FAILED (HR)) {

      FinalRes = S_FALSE;
      goto Exit;

    }

    FinalRes = S_OK;

  }

Exit:

  return FinalRes;

}





HRESULT SetPermissions (SE_OBJECT_TYPE const ObjectType, PCTSTR const ObjectPath, DWORD const Permissions) {

  HRESULT FinalRes;

  ACL * ACL = nullptr;

  {

    Trace (7, L"Setting our permissions %x for \"%s\"", Permissions, ObjectPath);

    HRESULT HR;

    Assert (OwnerToken != nullptr);

    EXPLICIT_ACCESS EA;
    ZeroObject (EA);

    EA.grfAccessPermissions = Permissions;
    EA.grfAccessMode = SET_ACCESS;
    EA.grfInheritance = NO_INHERITANCE;
    EA.Trustee.TrusteeForm = TRUSTEE_IS_SID;
    EA.Trustee.TrusteeType = TRUSTEE_IS_GROUP;
    EA.Trustee.ptstrName = reinterpret_cast <PTSTR> (OwnerToken->Owner);

    DWORD Res;

    Res = SetEntriesInAcl (1, &EA, nullptr, &ACL);

    if (Res != NO_ERROR) {

      Trace (9, L"SetEntriesInAcl failed (%s)", ErrorText (Res));

      FinalRes = HRESULT_FROM_WIN32 (Res);
      goto Exit;

    }

    bool OwnershipTaken = false;

    HR = S_OK;

    for (;;) {

      Res = SetNamedSecurityInfo (
        const_cast <PTSTR> (ObjectPath), ObjectType,
        DACL_SECURITY_INFORMATION, nullptr, nullptr,
        ACL, nullptr
      );

      if (Res == NO_ERROR) {

        Trace (8, L"Permissions for \"%s\" set to %x", ObjectPath, Permissions);

        break;

      }

      HR = HRESULT_FROM_WIN32 (Res);

      Trace (9, L"SetNamedSecurityInfo failed (%s)", ErrorText (HR));

      if (OwnershipTaken) {

        FinalRes = HR;
        goto Exit;

      }

      HR = TakeOwnership (ObjectType, ObjectPath);

      if (FAILED (HR)) {

        FinalRes = HR;
        goto Exit;

      }

      OwnershipTaken = true;

      Trace (7, L"Retrying to set permissions");

    } // retry loop

    Assert (Res == NO_ERROR && HR == S_OK);

    FinalRes = S_OK;

  }

Exit:

  if (ACL != nullptr) {

    LocalFree (ACL);
    ACL = nullptr;

  }

  return FinalRes;

}
